function [samptheta, samphyper, SAMPLEX, sampgammaAGM]= ode_ComGP_solveAGM(Data, ODE, GP, Iteration, GammaAGM,annealing)
%% ODE parameter estimation with common GP settings 


%%%%%%%%%%%%%%%%%% Inputs %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Data:
%      Data.y: observations
%      Data.samptime: the sample time

% ODE:
%      ODE.fun: ODE functions
%      ODE.num: the number of the parameters
%      ODE.discrete: the discretized ranges of the parameters
%      ODE.initial: initial values of ODE parameters
%      ODE.prior: prior for ODE parameters 

% GP:
%      GP.fun: GP covariance function: c(t,t')
%      GP.fun_d: GP derivative d{c(t,t')}/dt
%      GP.fun_dd: GP derivative d^2{c(t,t')}/dtdt'
%      GP.num: the number of the GP hyperparameters (we put the noise std to the end of the hyperparameter vector)
%      GP.discrete: the discretized ranges of the GP hyperparameters
%      GP.inital_X: initial mean, std of X, and the number of discretized bins for X
%      GP.inital: initial values of GP hyperparameters
%      GP.prior: prior for GP hyperparameters

% Iter:
%      Iteration.total: the number of the total iterations
%      Iteration.sub:   the number of the sub iterations

%%%%%%%%%%%%%%%%%%% Outputs %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% samptheta: samples of ODE parameters
% samphyper: samples of GP hyperparameters
% sampgammaAGM: samples of gradient mismatch parameter
% SAMPLEX: samples of X


%% Load Data and Sampletime
y=Data.y;
samptime=Data.samptime;

TTT=length(samptime);      % number of sampled points
meany=repmat(mean(y),TTT,1);    % observation mean

%% Discratized Grid
%ODE
for i=1:ODE.num
    ODE_par{i}=ODE.discrete{i};
end
%GP
for i=1:GP.num
    GP_par{i}=GP.discrete{i};
end

%% Initialization 
%ODE
odepar=ODE.initial;

%GP
MeanX=GP.initial_X{1};
StdX=GP.initial_X{2};
SX=MeanX;  %%% initial X as the mean of GP

gppar=GP.initial;

%%% gamma here is the gradient mismatch parameter
gammaAGM = GammaAGM.initial;

if annealing==0
for i=1:GammaAGM.num
    gammaAGM_par{i}=GammaAGM.discrete{i};
end
end

if annealing==1
	gammaAnnealed = gammaAGM;

	for i = 2:Iteration.total
		gammaAnnealed(i) = gammaAnnealed(i-1)*0.95;

	end
end

%% Gibbs Sampling

for iter=1:Iteration.total
    iter
    %##############  Sampling Parameters #################################   
    for subiter=1:Iteration.sub 
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%% Sample GP hyperparameters, fix ODE parameter, noise std and X
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %##### Note before sampling hyperparameters##############
        % Since X and ODE parameters are fixed, dotX are fixed
        for i=1:TTT
            dotx(i,:)=feval(ODE.fun,samptime(i),SX(i,:),odepar)';
        end    
        
        %#################################################################
        
        %%% Sample GP hyperparameters
        for k=1:GP.num-1             
            IT=GP_par{k}; % choose which hyperparameter will be sampled
            
            for j=1:length(IT)   
                gppar(k)=IT(j);
                
                for s_sampind=1:TTT
                    s=samptime(s_sampind);
                    for t_sampind=1:TTT
                        t=samptime(t_sampind);
                        Cxx(s_sampind,t_sampind)=feval(GP.fun,s,t,gppar);  
                        dC(s_sampind,t_sampind)=feval(GP.fun_d,s,t,gppar);                                             
                        ddC(s_sampind,t_sampind)=feval(GP.fun_dd,s,t,gppar);
                    end
                end
                
                Cxx=0.5*(Cxx+Cxx');  % for numerical stability, same throughout

                

			middleAGM = dC*inv(Cxx);

			mAGM = middleAGM*(SX-meany); 
			AAGM = ddC - middleAGM*(dC');

			AAGM = 0.5*(AAGM + AAGM');

                p1=1;
                for i=1:size(meany,2)
                    p1=p1*mvnpdf(dotx(:,i)',mAGM(:,i)',AAGM+(gammaAGM*eye(TTT)));
                end

                p2=1;
                for i=1:size(meany,2)
                    p2=p2*mvnpdf(y(:,i)',SX(:,i)',gppar(GP.num)^2*eye(TTT));
                end





                % p(x|hyperparameters)
                p3=1;
                for i=1:size(y,2)
                    p3=p3*mvnpdf(SX(:,i)',meany(:,i)',Cxx);  
                end
                
                % Prior
                p4=feval(GP.prior{k},gppar(k));

                pgp(j)=p1*p2*p3*p4;
            end
            
            ind = randsample(length(IT),1,true,pgp);          
            gppar(k)=IT(ind); 
            clear IT
            clear pgp
        end        
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%% Sample noise std, fix GP hyperparameters, ODE parameter and X
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %##### Note before sampling  noise std ###################
        
        % Since X and ODE parameters are fixed, dotX are still fixed as before
        
        % Since GP hyperparameters are fixed now, the covariance matrice are fixed
        
        for s_sampind=1:TTT
            s=samptime(s_sampind);
            for t_sampind=1:TTT
                t=samptime(t_sampind);
                Cxx(s_sampind,t_sampind)=feval(GP.fun,s,t,gppar);  
                dC(s_sampind,t_sampind)=feval(GP.fun_d,s,t,gppar);                                             
                ddC(s_sampind,t_sampind)=feval(GP.fun_dd,s,t,gppar);
            end
        end


       
        %#################################################################
                 
        % Sampling Sigma
        IT=GP_par{GP.num};
        for j=1:length(IT)
            gppar(GP.num)=IT(j);
            

           
			middleAGM = dC*inv(Cxx);

			mAGM = middleAGM*(SX-meany); 
			AAGM = ddC - middleAGM*(dC');

			AAGM = 0.5*(AAGM + AAGM');


                p1=1;
                for i=1:size(meany,2)
                    p1=p1*mvnpdf(dotx(:,i)',mAGM(:,i)',AAGM+(gammaAGM*eye(TTT)));
                end

                p2=1;
                for i=1:size(meany,2)
                    p2=p2*mvnpdf(y(:,i)',SX(:,i)',gppar(GP.num)^2*eye(TTT));
                end


            % Prior
            p3=feval(GP.prior{GP.num},gppar(GP.num)); 
            psg(j)=p1*p2*p3;
        end

        ind = randsample(length(IT),1,true,psg);
        gppar(GP.num)=IT(ind); 
        clear IT
        clear psg
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%% Sample ODE parameter, fix GP hyperparameters,  noise std and X
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        

			middleAGM = dC*inv(Cxx);

			mAGM = middleAGM*(SX-meany); 

			AAGM = ddC - middleAGM*(dC');

			AAGM = 0.5*(AAGM + AAGM');

        %#################################################
        
        %%% Sampling ODE parameters
        for k=1:ODE.num
            IT=ODE_par{k}; 
            for j=1:length(IT)
                odepar(k)=IT(j);
                
                for i=1:TTT
                     dotx(i,:)=feval(ODE.fun,samptime(i),SX(i,:),odepar)';
                     
                end             



                p1=1;
                for i=1:size(meany,2)
                    p1=p1*mvnpdf(dotx(:,i)',mAGM(:,i)',AAGM+(gammaAGM*eye(TTT)));
                end

                p2=1;
                for i=1:size(meany,2)
                    p2=p2*mvnpdf(y(:,i)',SX(:,i)',gppar(GP.num)^2*eye(TTT));
                end


                % Prior
                p3=feval(ODE.prior{k},odepar(k));
                
                pode(j)=p1*p2*p3;
            end

            ind = randsample(length(IT),1,true,pode);
            odepar(k)=IT(ind); 
            clear IT
            clear pode
        end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%% Sample slack parameter. The only variable that changes in the calculation of the conditional is the slack parameter, hence, calculate necessary components before the looping over different slack values
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if annealing == 0

        for s_sampind=1:TTT
                    s=samptime(s_sampind);
                    for t_sampind=1:TTT
                        t=samptime(t_sampind);
                        Cxx(s_sampind,t_sampind)=feval(GP.fun,s,t,gppar);  
                        dC(s_sampind,t_sampind)=feval(GP.fun_d,s,t,gppar);                                             
                        ddC(s_sampind,t_sampind)=feval(GP.fun_dd,s,t,gppar);
                    end
                end

                	for i=1:TTT
                     dotx(i,:)=feval(ODE.fun,samptime(i),SX				(i,:),odepar)';
                 end    

                Cxx=0.5*(Cxx+Cxx');  

                
			middleAGM = dC*inv(Cxx);

			mAGM = middleAGM*(SX-meany); 
			AAGM = ddC - middleAGM*(dC');

			AAGM = 0.5*(AAGM + AAGM');


%%%%%%%%%%%%%%%%%%%%

	for k=1:GammaAGM.num
		IT = gammaAGM_par{k};
		for j = 1:length(IT)
			gammaAGM(k) = IT(j);


	                p1=1;
                for i=1:size(meany,2)
                    p1=p1*mvnpdf(dotx(:,i)',mAGM(:,i)',AAGM+(gammaAGM*eye(TTT)));
                end


		p2 = feval(GammaAGM.prior{k},gammaAGM(k));

		pgammaAGM(j) = p1*p2;

	
	end

%%%%%%%%%%%%%%%%%%%%
	
	ind = randsample(length(IT),1,true,pgammaAGM);
	gammaAGM(k)=IT(ind);
	clear IT
	clear pslack
	end

    end % annealing == 0


    end % subiter

if annealing == 1
	gammaAGM = gammaAnnealed(iter);	
end

    samptheta(iter,:)=odepar;
    samphyper(iter,:)=gppar;
    sampgammaAGM(iter,:) = gammaAGM;
    
    %############# Sampling X along with dimension ########################
    
    %##### Note before sampling  X ###################      
        
    % Since GP hyperparameters and noise std are fixed,
              
   
     %#################################################
    			middleAGM = dC*inv(Cxx);



			AAGM = ddC - middleAGM*(dC');

			AAGM = 0.5*(AAGM + AAGM');


    for subiter=1:Iteration.sub
        
        for Tstep=1:TTT
            
            for DIM=1:size(y,2)
                
                 IT1=linspace(MeanX(Tstep,DIM)-3*StdX(Tstep,DIM),MeanX(Tstep,DIM)+3*StdX(Tstep,DIM),GP.initial_X{3});
        
                 for j=1:length(IT1)
                  
                     SX(Tstep,DIM)=IT1(j);
                     
                     for i=1:TTT
                         dotx(i,:)=feval(ODE.fun,samptime(i),SX(i,:),odepar)';
                         
                     end             

				mAGM = middleAGM*(SX-meany); %%% Check to make sure done for each species seperately %%%


                p1S=1;
                for i=1:size(meany,2)
                    p1S=p1S*mvnpdf(dotx(:,i)',mAGM(:,i)',AAGM+(gammaAGM*eye(TTT)));
                end

                p2S=1;
                for i=1:size(meany,2)
                    p2S=p2S*mvnpdf(y(:,i)',SX(:,i)',gppar(GP.num)^2*eye(TTT));
                end


                     
                     
                    % p(x|hyperparameters)
                     p3S=mvnpdf(SX(:,DIM)',meany(:,DIM)',Cxx);
                     px1(j)=p1S*p2S*p3S;
                 end
                 ind=randsample(length(IT1),1,true,px1);
                 SX(Tstep,DIM)=IT1(ind); 
                 clear IT1
                 clear px1
            end
            
        end
             
    end
    
    SAMPLEX{iter}=SX;
           
end        

